# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import json
import logging
import os
import pprint
import sys
import time
from graphlib import TopologicalSorter
from typing import Any
from typing import Dict
from typing import List

from taurus_datajob_api import ApiException
from vdk.plugin.meta_jobs.cached_data_job_executor import TrackingDataJobExecutor
from vdk.plugin.meta_jobs.meta import TrackableJob
from vdk.plugin.meta_jobs.meta_configuration import MetaPluginConfiguration
from vdk.plugin.meta_jobs.remote_data_job_executor import RemoteDataJobExecutor
from vdk.plugin.meta_jobs.time_based_queue import TimeBasedQueue

log = logging.getLogger(__name__)


class MetaJobsDag:
    def __init__(self, team_name: str, meta_config: MetaPluginConfiguration):
        self._team_name = team_name
        self._topological_sorter = TopologicalSorter()
        self._delayed_starting_jobs = TimeBasedQueue(
            min_ready_time_seconds=meta_config.meta_jobs_delayed_jobs_min_delay_seconds(),
            randomize_delay_seconds=meta_config.meta_jobs_delayed_jobs_randomized_added_delay_seconds(),
        )
        self._finished_jobs = []
        self._dag_execution_check_time_period_seconds = (
            meta_config.meta_jobs_dag_execution_check_time_period_seconds()
        )
        self._job_executor = TrackingDataJobExecutor(
            executor=RemoteDataJobExecutor(),
            time_between_status_check_seconds=meta_config.meta_jobs_time_between_status_check_seconds(),
        )

    def build_dag(self, jobs: List[Dict]):
        for job in jobs:
            # TODO: add some job validation here; check the job exists, its previous jobs exists, etc
            trackable_job = TrackableJob(
                job["job_name"],
                job.get("team_name", self._team_name),
                job.get("fail_meta_job_on_error", True),
            )
            self._job_executor.register_job(trackable_job)
            self._topological_sorter.add(trackable_job.job_name, *job["depends_on"])

    def execute_dag(self):
        self._topological_sorter.prepare()
        while self._topological_sorter.is_active():
            for node in self._topological_sorter.get_ready():
                self._start_job(node)
            self._start_delayed_jobs()

            finished_jobs = self._get_finished_jobs()
            self._finalize_jobs(finished_jobs)
            if not finished_jobs:
                # No jobs are finished at this iteration so let's wait a bit to let them
                # finish
                time.sleep(self._dag_execution_check_time_period_seconds)

    def _get_finished_jobs(self):
        return [
            job
            for job in self._job_executor.get_finished_job_names()
            if job not in self._finished_jobs
        ]

    def _finalize_jobs(self, finalized_jobs):
        for node in finalized_jobs:
            log.info(f"Data Job {node} has finished.")
            self._topological_sorter.done(node)
            self._job_executor.finalize_job(node)
            self._finished_jobs.append(node)

    def _start_delayed_jobs(self):
        while True:
            job = self._delayed_starting_jobs.dequeue()
            if job is None:
                break
            log.info(f"Trying to start job {job} again.")
            self._start_job(job)

    def __repr__(self):
        # TODO move out of this class
        def default_serialization(o: Any) -> Any:
            return o.__dict__ if "__dict__" in dir(o) else str(o)

        jobs = self._job_executor.get_all_jobs()
        data = [j.details if j.details is not None else j.__dict__ for j in jobs]
        try:
            result = json.dumps(data, default=default_serialization, indent=2)
        except Exception as e:
            log.debug(f"Failed to json.dumps : {e}. Fallback to pprint.")
            # sort_dicts is supported since 3.8
            if sys.version_info[0] >= 3 and sys.version_info[1] >= 8:
                result = pprint.pformat(
                    data, indent=2, depth=5, compact=False, sort_dicts=False
                )
            else:
                result = pprint.pformat(data, indent=2, depth=5, compact=False)
        return result

    def _start_job(self, node):
        try:
            self._job_executor.start_job(node)
        except ApiException as e:
            if e.status == 409:
                log.info(
                    f"Detected conflict with another running job: {e}. Will be re-tried later"
                )
                self._delayed_starting_jobs.enqueue(node)
            elif e.status >= 500:
                log.info(
                    f"Starting job fail with server error : {e}. Will be re-tried later"
                )
                self._delayed_starting_jobs.enqueue(node)
            else:
                raise



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import json
import logging
import os
import time
from typing import Dict
from typing import Optional

import urllib3.exceptions as url_exception
from vdk.internal.core.errors import ErrorMessage
from vdk.internal.core.errors import UserCodeError
from vdk.plugin.meta_jobs.meta import IDataJobExecutor
from vdk.plugin.meta_jobs.meta import TrackableJob
from vdk.plugin.meta_jobs.remote_data_job import JobStatus

log = logging.getLogger(__name__)

ACTIVE_JOB_STATUSES = [JobStatus.SUBMITTED.value, JobStatus.RUNNING.value]
SLEEP_TIME = 10
ALLOWED_RETRIES = 3


class TrackingDataJobExecutor:
    def __init__(
        self, executor: IDataJobExecutor, time_between_status_check_seconds: int
    ):
        self._executor = executor
        self._jobs_cache: Dict[str, TrackableJob] = dict()
        self._time_between_status_check_seconds = time_between_status_check_seconds

    def register_job(self, job: TrackableJob):
        if job.job_name in self._jobs_cache:
            log.warning(
                f"Job with name {job.job_name} aleady exists. Details: {self._jobs_cache[job.job_name]}. "
                f"Will overwrite it."
            )
        self._jobs_cache[job.job_name] = job

    def start_job(self, job_name: str) -> None:
        """
        :param job_name: the job to start and track
        """
        job = self.__get_job(job_name)
        job.start_attempt += 1
        execution_id = self.start_new_job_execution(
            job_name=job.job_name, team_name=job.team_name
        )
        log.info(f"Starting new data job execution with id {execution_id}")
        job.execution_id = execution_id
        job.status = JobStatus.SUBMITTED.value
        job.details = self._executor.details_job(
            job.job_name, job.team_name, job.execution_id
        )
        log.info(
            f"Started data job {job_name}:\n{self.__get_printable_details(job.details)}"
        )

    def finalize_job(self, job_name):
        job = self.__get_job(job_name)
        details = self._executor.details_job(
            job.job_name, job.team_name, job.execution_id
        )
        job.details = details
        log.info(
            f"Finished data job {job_name}:\n"
            f"start_time: {details['start_time']}\n"
            f"end_time: {details['end_time']}\n"
            f"status: {details['status']}\n"
            f"message: {details['message']}"
        )
        if job.status != JobStatus.SUCCEEDED.value and job.fail_meta_job_on_error:
            raise UserCodeError(
                ErrorMessage(
                    "",
                    "Meta Job failed due to a data job failure.",
                    f"Data Job {job_name} failed. See details: {details}",
                    "The rest of the jobs in the meta job will not be started "
                    "and the meta data job will fail.",
                    "Investigate the error in the job or re-try again.",
                )
            )

    @staticmethod
    def __get_printable_details(details):
        del details["deployment"]
        return json.dumps(details, default=lambda o: str(o), indent=2)

    def __get_job(self, job_name) -> TrackableJob:
        job: TrackableJob = self._jobs_cache.get(job_name)
        if job is None:
            raise IndexError(
                f"The job {job_name} has not been registered. Use register_job first. "
                "Alternatively, verify that all of the job names in your job list are spelled correctly."
            )
        return job

    @staticmethod
    def __is_job_submitted(job: TrackableJob):
        return job.status is not None

    def status(self, job_name: str) -> str:
        job = self.__get_job(job_name)
        if job.status in ACTIVE_JOB_STATUSES:
            job.status = self._executor.status_job(
                job.job_name, job.team_name, job.execution_id
            )
        log.debug(f"Job status: {job}")
        return job.status

    def get_finished_job_names(self):
        finalized_jobs = []
        # TODO: optimize
        # Do not call the status every time (use TTL caching)
        # Do not call all status at the same time - stagger them in time
        # Or use GraphQL API to get status at once (batch)
        for job in self._jobs_cache.values():
            if (
                self.__is_job_submitted(job)
                and job.status in ACTIVE_JOB_STATUSES
                and time.time() - job.last_status_time
                > self._time_between_status_check_seconds
            ):
                job.last_status_time = time.time()
                job.status = self.status(job.job_name)

        for job in self._jobs_cache.values():
            if self.__is_job_submitted(job) and job.status not in ACTIVE_JOB_STATUSES:
                finalized_jobs.append(job.job_name)
        return finalized_jobs

    def get_all_jobs(self):
        return list(self._jobs_cache.values())

    def get_currently_running_jobs(self):
        return [j for j in self._jobs_cache.values() if j.status in ACTIVE_JOB_STATUSES]

    def start_new_job_execution(self, job_name: str, team_name: str) -> str:
        """
        Start a new data job execution.
        The stages of the process are:
            1) Get the latest available execution_id before any new
               executions have been started.
            2) Start a new execution.
            3) In case of a Timeout exception, check if a new execution
               has been started and re-try step 2 if needed.

        NOTE: A loop is used to handle situations, where due to some
        network instability, a socket timeout happens. The execution of
        the data job may have started, but we don't know because an execution
        id was not returned due to the exception.

        :param job_name: name of the data job to be executed
        :param team_name: name of the owning team
        :return: id of the started job execution
        """
        current_retries = 0
        execution_id = None

        latest_available_execution_id = self.get_latest_available_execution_id(
            job_name=job_name, team=team_name
        )

        while current_retries < ALLOWED_RETRIES:
            try:
                execution_id = self._executor.start_job(job_name, team_name)
                return execution_id
            except url_exception.TimeoutError as e:
                log.info(
                    f"A timeout exception occurred while starting the {job_name} data job. "
                    f"Exception was: {e}"
                )

                execution_id = self.get_latest_available_execution_id(
                    job_name=job_name, team=team_name
                )
                if execution_id and execution_id != latest_available_execution_id:
                    return execution_id
                else:
                    current_retries += 1
                    time.sleep(SLEEP_TIME)  # Sleep for 10 seconds before re-try.

        # If the execution reaches this point, then something has happened,
        # and the state of the data job that has been started cannot be determined.
        raise RuntimeError(
            "Bug, fix me! Something wrong has happened and I cannot recover!"
        )

    def get_latest_available_execution_id(
        self, job_name: str, team: str
    ) -> Optional[str]:
        """
        Get the latest execution_id from the list of all executions for a
        given data job.

        :param job_name: name of data job for which an execution_id is needed
        :param team: name of the team owning the data job
        :return: The execution_id of the latest job execution or None if no
                 executions are available.
        """
        latest_execution_id = None
        executions_list = self._executor.job_executions_list(
            job_name=job_name, team_name=team
        )

        if executions_list:
            # We need only the latest execution which is the last element of the list
            latest_execution_id = executions_list[-1].id
        return latest_execution_id



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from unittest.mock import MagicMock
from unittest.mock import patch

from taurus_datajob_api import DataJobDeployment
from taurus_datajob_api import DataJobExecution
from urllib3.exceptions import ReadTimeoutError
from vdk.plugin.meta_jobs.cached_data_job_executor import TrackingDataJobExecutor
from vdk.plugin.meta_jobs.remote_data_job_executor import RemoteDataJobExecutor


def test_get_latest_available_execution_id():
    test_job_id = "awesome-test_job"
    test_execution: DataJobExecution = DataJobExecution(
        id=test_job_id,
        job_name=test_job_id,
        logs_url="http://url",
        deployment=DataJobDeployment(),
        start_time="2021-09-24T14:14:03.922Z",
        status="succeeded",
        message="foo",
    )

    test_executor = MagicMock(spec=RemoteDataJobExecutor)
    test_executor.job_executions_list.side_effect = [[test_execution]]

    tracking_executor = TrackingDataJobExecutor(test_executor, 40)
    assert test_job_id == tracking_executor.get_latest_available_execution_id(
        test_job_id, "awesome-test-team"
    )


def test_get_latest_available_execution_id_return_none():
    test_job_id = "awesome-test-job"

    test_executor = MagicMock(spec=RemoteDataJobExecutor)
    test_executor.job_executions_list.side_effect = [[]]

    tracking_executor = TrackingDataJobExecutor(test_executor, 40)
    assert not tracking_executor.get_latest_available_execution_id(
        test_job_id, "awesome-test-team"
    )


@patch("vdk.plugin.meta_jobs.cached_data_job_executor.SLEEP_TIME", return_value=1)
def test_start_new_job_execution_timeout_error(patched_timeout):
    test_job_id = "awesome-test-job"
    test_timeout_exception = ReadTimeoutError(
        message="Read timed out.", url="http://url", pool=MagicMock()
    )

    test_executor = MagicMock(spec=RemoteDataJobExecutor)
    test_executor.start_job.side_effect = [test_timeout_exception, "awesome-test-job"]
    test_executor.job_executions_list.side_effect = [[], []]

    tracking_executor = TrackingDataJobExecutor(test_executor, 40)
    assert test_job_id == tracking_executor.start_new_job_execution(
        job_name=test_job_id, team_name="awesome-team"
    )
    assert len(test_executor.method_calls) == 4



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import abc
from dataclasses import dataclass
from typing import List
from typing import Optional

from taurus_datajob_api import DataJobExecution
from vdk.plugin.meta_jobs.api import meta_job


class IDataJobExecutor(abc.ABC):
    @abc.abstractmethod
    def start_job(self, job_name: str, team_name: str):
        """
        Start an execution of a data job and returns its execution id
        :param job_name:
        :param team_name:
        :return: execution id
        """
        pass

    @abc.abstractmethod
    def status_job(self, job_name: str, team_name: str, execution_id: str):
        """
        Get the current status of a data job execution
        :param job_name:
        :param team_name:
        :param execution_id:
        :return: status in string as defined by Control Service API
        """
        pass

    @abc.abstractmethod
    def details_job(self, job_name: str, team_name: str, execution_id: str) -> dict:
        """
        Get the current status of a data job execution
        :param job_name:
        :param team_name:
        :param execution_id:
        :return: status in string as defined by Control Service API
        """
        pass

    @abc.abstractmethod
    def job_executions_list(
        self, job_name: str, team_name: str
    ) -> Optional[List[DataJobExecution]]:
        """
        Get the list of all recent executions of a data job. The list returned by
        the Control Srvice is sorted, and the lat element is the latest execution.
        :param job_name: Name of data job for which executions are to be returned.
        :param team_name: Name of the team owning the data job.
        :return: A list of DataJobExecution objects.
        """
        pass


@dataclass
class TrackableJob(meta_job.SingleJob):
    status: str = None
    execution_id: str = None
    details: dict = None
    start_attempt = 0
    last_status_time = 0



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import time
from unittest.mock import call
from unittest.mock import MagicMock

from vdk.plugin.meta_jobs.cached_data_job_executor import TrackingDataJobExecutor
from vdk.plugin.meta_jobs.meta_dag import MetaJobsDag

# We overall eschew unit tests in favor of functional tests in test_meta_job
# Still some functionalities are more easily tested in unit tests so we add here some.


class DummyMetaPluginConfiguration:
    def meta_jobs_delayed_jobs_min_delay_seconds(self):
        return 30

    def meta_jobs_delayed_jobs_randomized_added_delay_seconds(self):
        return 600

    def meta_jobs_dag_execution_check_time_period_seconds(self):
        return 10

    def meta_jobs_time_between_status_check_seconds(self):
        return 40


def test_execute_dag_happy_case():
    job1 = dict(job_name="job1", depends_on=[])
    job2 = dict(job_name="job2", depends_on=["job1"])
    job3 = dict(job_name="job3", depends_on=["job1"])
    job4 = dict(job_name="job4", depends_on=["job2", "job3"])
    jobs = [job1, job2, job3, job4]

    dag = MetaJobsDag("team", DummyMetaPluginConfiguration())
    dag.build_dag(jobs)
    dag._job_executor = MagicMock(spec=TrackingDataJobExecutor)
    dag._job_executor.get_finished_job_names.side_effect = [
        ["job1"],
        ["job2", "job3"],
        ["job4"],
    ]

    dag.execute_dag()

    assert [
        call("job1"),
        call("job2"),
        call("job3"),
        call("job4"),
    ] == dag._job_executor.start_job.call_args_list


def test_execute_dag_busyloop():
    job1 = dict(job_name="job1", depends_on=[])
    job2 = dict(job_name="job2", depends_on=["job1"])
    job3 = dict(job_name="job3", depends_on=["job1"])
    jobs = [job1, job2, job3]

    dag = MetaJobsDag("team", DummyMetaPluginConfiguration())
    dag.build_dag(jobs)
    dag._job_executor = MagicMock(spec=TrackingDataJobExecutor)
    dag._dag_execution_check_time_period_seconds = 3

    calls = [0]
    start_time = [0]

    def mock_get_finished_job_names(*args, **kwargs):
        calls[0] += 1
        if calls[0] == 1:
            start_time[0] = time.time()
            return ["job1"]
        elif time.time() - start_time[0] > 2:
            return ["job2", "job3"]
        else:
            return []

    dag._job_executor.get_finished_job_names.side_effect = mock_get_finished_job_names

    dag.execute_dag()

    # check for busyloop (this would have been called hundreds of times if there is busyloop bug)
    assert calls[0] <= 4



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import time

from vdk.plugin.meta_jobs.time_based_queue import TimeBasedQueue


def test_enqueue_dequeue_empty():
    queue = TimeBasedQueue(
        min_ready_time_seconds=2, randomize_delay_seconds=0, dequeue_timeout_seconds=0
    )
    assert queue.dequeue() is None


def test_enqueue_dequeue_not_wait_for_ready():
    queue = TimeBasedQueue(
        min_ready_time_seconds=4, randomize_delay_seconds=0, dequeue_timeout_seconds=0
    )
    queue.enqueue(1)
    queue.enqueue(2)
    assert queue.dequeue() is None


def test_enqueue_dequeue_wait():
    queue = TimeBasedQueue(
        min_ready_time_seconds=2, randomize_delay_seconds=0, dequeue_timeout_seconds=0
    )
    queue.enqueue(1)
    queue.enqueue(2)
    assert queue.dequeue() is None
    time.sleep(2)
    assert queue.dequeue() == 1
    assert queue.dequeue() == 2



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import pathlib

import setuptools

"""
Builds a package with the help of setuptools in order for this package to be imported in other projects
"""

__version__ = "0.1.802837202"

setuptools.setup(
    name="vdk-meta-jobs",
    version=__version__,
    url="https://github.com/vmware/versatile-data-kit",
    description="Express dependecies between data jobs.",
    long_description=pathlib.Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    install_requires=[
        "vdk-core",
        "graphlib-backport",
        "vdk-control-api-auth",
        "vdk-plugin-control-cli",
        "vdk-control-service-api",
        "urllib3",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_namespace_packages(where="src"),
    # This is the only vdk plugin specifc part
    # Define entry point called "vdk.plugin.run" with name of plugin and module to act as entry point.
    entry_points={"vdk.plugin.run": ["meta-jobs = vdk.plugin.meta_jobs.plugin_entry"]},
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import json
import os
import time
from unittest import mock

from click.testing import Result
from pytest_httpserver.pytest_plugin import PluginHTTPServer
from taurus_datajob_api import DataJobDeployment
from taurus_datajob_api import DataJobExecution
from vdk.internal.core.errors import UserCodeError
from vdk.plugin.meta_jobs import plugin_entry
from vdk.plugin.test_utils.util_funcs import cli_assert_equal
from vdk.plugin.test_utils.util_funcs import CliEntryBasedTestRunner
from vdk.plugin.test_utils.util_funcs import jobs_path_from_caller_directory
from werkzeug import Request
from werkzeug import Response


def _prepare(httpserver: PluginHTTPServer, jobs=None):
    """
    :param httpserver: the pytest http server
    :param jobs: list of jobs in format ('job-name', [list of http statuses to be returned], job result status)
    :return:
    """
    rest_api_url = httpserver.url_for("")
    team_name = "team-awesome"
    if jobs is None:
        jobs = [
            ("job1", [200], "succeeded", 0),
            ("job2", [200], "succeeded", 0),
            ("job3", [200], "succeeded", 0),
            ("job4", [200], "succeeded", 0),
        ]

    started_jobs = dict()

    for job_name, request_responses, job_status, *execution_duration in jobs:
        request_responses.reverse()
        execution_duration = execution_duration[0] if execution_duration else 0

        def handler(location, statuses, job_name):
            def _handler_fn(r: Request):
                status = statuses[0] if len(statuses) == 1 else statuses.pop()
                if status < 300:
                    started_jobs[job_name] = time.time()
                return Response(status=status, headers=dict(Location=location))

            return _handler_fn

        httpserver.expect_request(
            uri=f"/data-jobs/for-team/{team_name}/jobs/{job_name}/deployments/production/executions",
            method="POST",
        ).respond_with_handler(
            handler(
                f"/data-jobs/for-team/{team_name}/jobs/{job_name}/executions/{job_name}",
                request_responses,
                job_name,
            )
        )

        def exec_handler(job_name, job_status, execution_duration):
            def _handler_fn(r: Request):
                actual_job_status = job_status
                if time.time() < started_jobs.get(job_name, 0) + execution_duration:
                    actual_job_status = "running"
                execution: DataJobExecution = DataJobExecution(
                    id=job_name,
                    job_name=job_name,
                    logs_url="http://url",
                    deployment=DataJobDeployment(),
                    start_time="2021-09-24T14:14:03.922Z",
                    status=actual_job_status,
                    message="foo",
                )
                response_data = json.dumps(execution.to_dict(), indent=4)
                return Response(
                    response_data,
                    status=200,
                    headers=None,
                    content_type="application/json",
                )

            return _handler_fn

        httpserver.expect_request(
            uri=f"/data-jobs/for-team/{team_name}/jobs/{job_name}/executions/{job_name}",
            method="GET",
        ).respond_with_handler(exec_handler(job_name, job_status, execution_duration))

        def exec_list_handler(job_name):
            def _handler_fn(r: Request):
                execution: DataJobExecution = DataJobExecution(
                    id=f"{job_name}-latest",
                    job_name=job_name,
                    logs_url="http://url",
                    deployment=DataJobDeployment(),
                    start_time="2021-09-24T14:14:03.922Z",
                    status="succeeded",
                    message="foo",
                )
                response_data = json.dumps(execution.to_dict(), indent=4)
                return Response(
                    [response_data],
                    status=200,
                    headers=None,
                    content_type="application/json",
                )

            return _handler_fn

        httpserver.expect_request(
            uri=f"/data-jobs/for-team/{team_name}/jobs/{job_name}/executions",
            method="GET",
        ).respond_with_handler(exec_list_handler(job_name))

    return rest_api_url


def test_meta_job(httpserver: PluginHTTPServer):
    api_url = _prepare(httpserver)

    with mock.patch.dict(
        os.environ,
        {"VDK_CONTROL_SERVICE_REST_API_URL": api_url},
    ):
        # CliEntryBasedTestRunner (provided by vdk-test-utils) gives a away to simulate vdk command
        # and mock large parts of it - e.g passed our own plugins
        runner = CliEntryBasedTestRunner(plugin_entry)

        result: Result = runner.invoke(
            ["run", jobs_path_from_caller_directory("meta-job")]
        )
        cli_assert_equal(0, result)


def test_meta_job_error(httpserver: PluginHTTPServer):
    jobs = [
        ("job1", [200], "succeeded"),
        ("job2", [200], "succeeded"),
        ("job3", [200], "platform_error"),
        ("job4", [200], "succeeded"),
    ]
    api_url = _prepare(httpserver, jobs)

    with mock.patch.dict(
        os.environ,
        {"VDK_CONTROL_SERVICE_REST_API_URL": api_url},
    ):
        # CliEntryBasedTestRunner (provided by vdk-test-utils) gives a away to simulate vdk command
        # and mock large parts of it - e.g passed our own plugins
        runner = CliEntryBasedTestRunner(plugin_entry)

        result: Result = runner.invoke(
            ["run", jobs_path_from_caller_directory("meta-job")]
        )
        cli_assert_equal(1, result)


def test_meta_job_fail_false(httpserver: PluginHTTPServer):
    jobs = [
        ("job1", [200], "succeeded"),
        ("job2", [200], "platform_error"),
        ("job3", [200], "succeeded"),
        ("job4", [200], "succeeded"),
    ]
    api_url = _prepare(httpserver, jobs)

    with mock.patch.dict(
        os.environ,
        {"VDK_CONTROL_SERVICE_REST_API_URL": api_url},
    ):
        # CliEntryBasedTestRunner (provided by vdk-test-utils) gives a away to simulate vdk command
        # and mock large parts of it - e.g passed our own plugins
        runner = CliEntryBasedTestRunner(plugin_entry)

        result: Result = runner.invoke(
            ["run", jobs_path_from_caller_directory("meta-job")]
        )
        cli_assert_equal(0, result)


def test_meta_job_conflict(httpserver: PluginHTTPServer):
    jobs = [
        ("job1", [409, 200], "succeeded"),
        ("job2", [500, 200], "succeeded"),
        ("job3", [200], "succeeded"),
        ("job4", [200], "succeeded"),
    ]
    api_url = _prepare(httpserver, jobs)

    with mock.patch.dict(
        os.environ,
        {
            "VDK_CONTROL_SERVICE_REST_API_URL": api_url,
            "VDK_META_JOBS_DELAYED_JOBS_RANDOMIZED_ADDED_DELAY_SECONDS": "0",
            "VDK_META_JOBS_DELAYED_JOBS_MIN_DELAY_SECONDS": "0",
        },
    ):
        # CliEntryBasedTestRunner (provided by vdk-test-utils) gives a away to simulate vdk command
        # and mock large parts of it - e.g passed our own plugins
        runner = CliEntryBasedTestRunner(plugin_entry)

        result: Result = runner.invoke(
            ["run", jobs_path_from_caller_directory("meta-job")]
        )
        cli_assert_equal(0, result)


def test_meta_job_cannot_start_job(httpserver: PluginHTTPServer):
    jobs = [
        ("job1", [401, 200], "succeeded"),
        ("job2", [200], "succeeded"),
        ("job3", [200], "succeeded"),
        ("job4", [200], "succeeded"),
    ]
    api_url = _prepare(httpserver, jobs)

    with mock.patch.dict(
        os.environ,
        {"VDK_CONTROL_SERVICE_REST_API_URL": api_url},
    ):
        # CliEntryBasedTestRunner (provided by vdk-test-utils) gives a away to simulate vdk command
        # and mock large parts of it - e.g passed our own plugins
        runner = CliEntryBasedTestRunner(plugin_entry)

        result: Result = runner.invoke(
            ["run", jobs_path_from_caller_directory("meta-job")]
        )
        cli_assert_equal(1, result)
        # we should have 2 requests in the log, one to get a list
        # of all executions, and one for the failing data job
        # no other request should be tried as the meta job fails
        assert len(httpserver.log) == 2


def test_meta_job_long_running(httpserver: PluginHTTPServer):
    jobs = [
        ("job1", [200], "succeeded", 3),  # execution duration is 3 seconds
        ("job2", [200], "succeeded"),
        ("job3", [200], "succeeded"),
        ("job4", [200], "succeeded"),
    ]
    api_url = _prepare(httpserver, jobs)

    with mock.patch.dict(
        os.environ,
        {
            "VDK_CONTROL_SERVICE_REST_API_URL": api_url,
            # we set 5 seconds more than execution duration of 3 set above
            "VDK_META_JOBS_TIME_BETWEEN_STATUS_CHECK_SECONDS": "5",
            "VDK_META_JOBS_DAG_EXECUTION_CHECK_TIME_PERIOD_SECONDS": "0",
        },
    ):
        # CliEntryBasedTestRunner (provided by vdk-test-utils) gives a away to simulate vdk command
        # and mock large parts of it - e.g passed our own plugins
        runner = CliEntryBasedTestRunner(plugin_entry)

        result: Result = runner.invoke(
            ["run", jobs_path_from_caller_directory("meta-job")]
        )
        cli_assert_equal(0, result)
        job1_requests = [
            req
            for req, res in httpserver.log
            if req.method == "GET" and req.base_url.endswith("job1")
        ]
        # We have 1 call during start, 1 call at finish and 1 call that returns running and 1 that returns the final
        # status. For total of 4
        # NB: test (verification) that requires that deep implementation details knowledge is
        # not a good idea but we need to verify that we are not hammering the API Server somehow ...
        assert len(job1_requests) == 4

        # let's make sure something else is not generating more requests then expected
        # if implementation is changed the number below would likely change.
        # If the new count is not that big we can edit it here to pass the test,
        # if the new count is too big, we have an issue that need to be investigated.
        assert len(httpserver.log) == 21


def test_meta_job_circular_dependency(httpserver: PluginHTTPServer):
    jobs = [
        ("job1", [200], "succeeded"),
        ("job2", [200], "succeeded"),
        ("job3", [200], "succeeded"),
        ("job4", [200], "succeeded"),
    ]
    api_url = _prepare(httpserver, jobs)

    with mock.patch.dict(
        os.environ,
        {"VDK_CONTROL_SERVICE_REST_API_URL": api_url},
    ):
        # CliEntryBasedTestRunner (provided by vdk-test-utils) gives a away to simulate vdk command
        # and mock large parts of it - e.g passed our own plugins
        runner = CliEntryBasedTestRunner(plugin_entry)

        result: Result = runner.invoke(
            ["run", jobs_path_from_caller_directory("meta-job-circular-dep")]
        )
        cli_assert_equal(1, result)
        # no other request should be tried as the meta job fails
        assert isinstance(result.exception, UserCodeError)
        assert len(httpserver.log) == 0
